#include <aarch64/mmu.h>

/**
 * The layout of physical memory space of virt:
 * 
 * 0..128MB is space for a flash device so we can run bootrom code such as UEFI.
 * 128MB..256MB is used for miscellaneous device I/O.
 * 256MB..1GB is reserved for possible future PCI support (ie where the
 * PCI memory window will go if we add a PCI host controller).
 * 1GB and up is RAM (which may happily spill over into the
 * high memory region beyond 4GB).
 * This represents a compromise between how much RAM can be given to
 * a 32 bit VM and leaving space for expansion and in particular for PCI.
 * Note that devices should generally be placed at multiples of 0x10000,
 * to accommodate guests using 64K pages.
 */

/**
 * Bits used to index: 29:21
 * Size per entry: 2MB
 * Address Span: [0x0, 0x40000000 (1GB)]
 */
__attribute__((__aligned__(PAGE_SIZE))) PTEntries _kernel_pt_lv2_dev = {
    // Space up to 0x8000000 is reserved for a boot ROM
    0x0 & ~PTE_VALID,
    0x200000 & ~PTE_VALID,
    0x400000 & ~PTE_VALID,
    0x600000 & ~PTE_VALID,
    0x800000 & ~PTE_VALID,
    0xa00000 & ~PTE_VALID,
    0xc00000 & ~PTE_VALID,
    0xe00000 & ~PTE_VALID,
    0x1000000 & ~PTE_VALID,
    0x1200000 & ~PTE_VALID,
    0x1400000 & ~PTE_VALID,
    0x1600000 & ~PTE_VALID,
    0x1800000 & ~PTE_VALID,
    0x1a00000 & ~PTE_VALID,
    0x1c00000 & ~PTE_VALID,
    0x1e00000 & ~PTE_VALID,
    0x2000000 & ~PTE_VALID,
    0x2200000 & ~PTE_VALID,
    0x2400000 & ~PTE_VALID,
    0x2600000 & ~PTE_VALID,
    0x2800000 & ~PTE_VALID,
    0x2a00000 & ~PTE_VALID,
    0x2c00000 & ~PTE_VALID,
    0x2e00000 & ~PTE_VALID,
    0x3000000 & ~PTE_VALID,
    0x3200000 & ~PTE_VALID,
    0x3400000 & ~PTE_VALID,
    0x3600000 & ~PTE_VALID,
    0x3800000 & ~PTE_VALID,
    0x3a00000 & ~PTE_VALID,
    0x3c00000 & ~PTE_VALID,
    0x3e00000 & ~PTE_VALID,
    0x4000000 & ~PTE_VALID,
    0x4200000 & ~PTE_VALID,
    0x4400000 & ~PTE_VALID,
    0x4600000 & ~PTE_VALID,
    0x4800000 & ~PTE_VALID,
    0x4a00000 & ~PTE_VALID,
    0x4c00000 & ~PTE_VALID,
    0x4e00000 & ~PTE_VALID,
    0x5000000 & ~PTE_VALID,
    0x5200000 & ~PTE_VALID,
    0x5400000 & ~PTE_VALID,
    0x5600000 & ~PTE_VALID,
    0x5800000 & ~PTE_VALID,
    0x5a00000 & ~PTE_VALID,
    0x5c00000 & ~PTE_VALID,
    0x5e00000 & ~PTE_VALID,
    0x6000000 & ~PTE_VALID,
    0x6200000 & ~PTE_VALID,
    0x6400000 & ~PTE_VALID,
    0x6600000 & ~PTE_VALID,
    0x6800000 & ~PTE_VALID,
    0x6a00000 & ~PTE_VALID,
    0x6c00000 & ~PTE_VALID,
    0x6e00000 & ~PTE_VALID,
    0x7000000 & ~PTE_VALID,
    0x7200000 & ~PTE_VALID,
    0x7400000 & ~PTE_VALID,
    0x7600000 & ~PTE_VALID,
    0x7800000 & ~PTE_VALID,
    0x7a00000 & ~PTE_VALID,
    0x7c00000 & ~PTE_VALID,
    0x7e00000 & ~PTE_VALID,

    // GIC
    0x8000000 | PTE_KERNEL_DEVICE,
    0x8200000 | PTE_KERNEL_DEVICE,
    0x8400000 | PTE_KERNEL_DEVICE,
    0x8600000 | PTE_KERNEL_DEVICE,
    0x8800000 | PTE_KERNEL_DEVICE,
    0x8a00000 | PTE_KERNEL_DEVICE,
    0x8c00000 | PTE_KERNEL_DEVICE,
    0x8e00000 | PTE_KERNEL_DEVICE,

    // UART0
    0x9000000 | PTE_KERNEL_DEVICE,
    0x9200000 | PTE_KERNEL_DEVICE,
    0x9400000 | PTE_KERNEL_DEVICE,
    0x9600000 | PTE_KERNEL_DEVICE,
    0x9800000 | PTE_KERNEL_DEVICE,
    0x9a00000 | PTE_KERNEL_DEVICE,
    0x9c00000 | PTE_KERNEL_DEVICE,
    0x9e00000 | PTE_KERNEL_DEVICE,

    // VIRTIO
    0xa000000 | PTE_KERNEL_DEVICE,
    0xa200000 | PTE_KERNEL_DEVICE,
    0xa400000 | PTE_KERNEL_DEVICE,
    0xa600000 | PTE_KERNEL_DEVICE,
    0xa800000 | PTE_KERNEL_DEVICE,
    0xaa00000 | PTE_KERNEL_DEVICE,
    0xac00000 | PTE_KERNEL_DEVICE,
    0xae00000 | PTE_KERNEL_DEVICE,
};

/**
 * Bits used to index: 29:21
 * Size per entry: 2MB
 * Address Span: [0x40000000 (1GB), 0x80000000 (2GB)]
 */
__attribute__((__aligned__(PAGE_SIZE))) PTEntries _kernel_pt_lv2_ram = {
    0x40000000 | PTE_KERNEL_DATA, 0x40200000 | PTE_KERNEL_DATA,
    0x40400000 | PTE_KERNEL_DATA, 0x40600000 | PTE_KERNEL_DATA,
    0x40800000 | PTE_KERNEL_DATA, 0x40a00000 | PTE_KERNEL_DATA,
    0x40c00000 | PTE_KERNEL_DATA, 0x40e00000 | PTE_KERNEL_DATA,
    0x41000000 | PTE_KERNEL_DATA, 0x41200000 | PTE_KERNEL_DATA,
    0x41400000 | PTE_KERNEL_DATA, 0x41600000 | PTE_KERNEL_DATA,
    0x41800000 | PTE_KERNEL_DATA, 0x41a00000 | PTE_KERNEL_DATA,
    0x41c00000 | PTE_KERNEL_DATA, 0x41e00000 | PTE_KERNEL_DATA,
    0x42000000 | PTE_KERNEL_DATA, 0x42200000 | PTE_KERNEL_DATA,
    0x42400000 | PTE_KERNEL_DATA, 0x42600000 | PTE_KERNEL_DATA,
    0x42800000 | PTE_KERNEL_DATA, 0x42a00000 | PTE_KERNEL_DATA,
    0x42c00000 | PTE_KERNEL_DATA, 0x42e00000 | PTE_KERNEL_DATA,
    0x43000000 | PTE_KERNEL_DATA, 0x43200000 | PTE_KERNEL_DATA,
    0x43400000 | PTE_KERNEL_DATA, 0x43600000 | PTE_KERNEL_DATA,
    0x43800000 | PTE_KERNEL_DATA, 0x43a00000 | PTE_KERNEL_DATA,
    0x43c00000 | PTE_KERNEL_DATA, 0x43e00000 | PTE_KERNEL_DATA,
    0x44000000 | PTE_KERNEL_DATA, 0x44200000 | PTE_KERNEL_DATA,
    0x44400000 | PTE_KERNEL_DATA, 0x44600000 | PTE_KERNEL_DATA,
    0x44800000 | PTE_KERNEL_DATA, 0x44a00000 | PTE_KERNEL_DATA,
    0x44c00000 | PTE_KERNEL_DATA, 0x44e00000 | PTE_KERNEL_DATA,
    0x45000000 | PTE_KERNEL_DATA, 0x45200000 | PTE_KERNEL_DATA,
    0x45400000 | PTE_KERNEL_DATA, 0x45600000 | PTE_KERNEL_DATA,
    0x45800000 | PTE_KERNEL_DATA, 0x45a00000 | PTE_KERNEL_DATA,
    0x45c00000 | PTE_KERNEL_DATA, 0x45e00000 | PTE_KERNEL_DATA,
    0x46000000 | PTE_KERNEL_DATA, 0x46200000 | PTE_KERNEL_DATA,
    0x46400000 | PTE_KERNEL_DATA, 0x46600000 | PTE_KERNEL_DATA,
    0x46800000 | PTE_KERNEL_DATA, 0x46a00000 | PTE_KERNEL_DATA,
    0x46c00000 | PTE_KERNEL_DATA, 0x46e00000 | PTE_KERNEL_DATA,
    0x47000000 | PTE_KERNEL_DATA, 0x47200000 | PTE_KERNEL_DATA,
    0x47400000 | PTE_KERNEL_DATA, 0x47600000 | PTE_KERNEL_DATA,
    0x47800000 | PTE_KERNEL_DATA, 0x47a00000 | PTE_KERNEL_DATA,
    0x47c00000 | PTE_KERNEL_DATA, 0x47e00000 | PTE_KERNEL_DATA,
    0x48000000 | PTE_KERNEL_DATA, 0x48200000 | PTE_KERNEL_DATA,
    0x48400000 | PTE_KERNEL_DATA, 0x48600000 | PTE_KERNEL_DATA,
    0x48800000 | PTE_KERNEL_DATA, 0x48a00000 | PTE_KERNEL_DATA,
    0x48c00000 | PTE_KERNEL_DATA, 0x48e00000 | PTE_KERNEL_DATA,
    0x49000000 | PTE_KERNEL_DATA, 0x49200000 | PTE_KERNEL_DATA,
    0x49400000 | PTE_KERNEL_DATA, 0x49600000 | PTE_KERNEL_DATA,
    0x49800000 | PTE_KERNEL_DATA, 0x49a00000 | PTE_KERNEL_DATA,
    0x49c00000 | PTE_KERNEL_DATA, 0x49e00000 | PTE_KERNEL_DATA,
    0x4a000000 | PTE_KERNEL_DATA, 0x4a200000 | PTE_KERNEL_DATA,
    0x4a400000 | PTE_KERNEL_DATA, 0x4a600000 | PTE_KERNEL_DATA,
    0x4a800000 | PTE_KERNEL_DATA, 0x4aa00000 | PTE_KERNEL_DATA,
    0x4ac00000 | PTE_KERNEL_DATA, 0x4ae00000 | PTE_KERNEL_DATA,
    0x4b000000 | PTE_KERNEL_DATA, 0x4b200000 | PTE_KERNEL_DATA,
    0x4b400000 | PTE_KERNEL_DATA, 0x4b600000 | PTE_KERNEL_DATA,
    0x4b800000 | PTE_KERNEL_DATA, 0x4ba00000 | PTE_KERNEL_DATA,
    0x4bc00000 | PTE_KERNEL_DATA, 0x4be00000 | PTE_KERNEL_DATA,
    0x4c000000 | PTE_KERNEL_DATA, 0x4c200000 | PTE_KERNEL_DATA,
    0x4c400000 | PTE_KERNEL_DATA, 0x4c600000 | PTE_KERNEL_DATA,
    0x4c800000 | PTE_KERNEL_DATA, 0x4ca00000 | PTE_KERNEL_DATA,
    0x4cc00000 | PTE_KERNEL_DATA, 0x4ce00000 | PTE_KERNEL_DATA,
    0x4d000000 | PTE_KERNEL_DATA, 0x4d200000 | PTE_KERNEL_DATA,
    0x4d400000 | PTE_KERNEL_DATA, 0x4d600000 | PTE_KERNEL_DATA,
    0x4d800000 | PTE_KERNEL_DATA, 0x4da00000 | PTE_KERNEL_DATA,
    0x4dc00000 | PTE_KERNEL_DATA, 0x4de00000 | PTE_KERNEL_DATA,
    0x4e000000 | PTE_KERNEL_DATA, 0x4e200000 | PTE_KERNEL_DATA,
    0x4e400000 | PTE_KERNEL_DATA, 0x4e600000 | PTE_KERNEL_DATA,
    0x4e800000 | PTE_KERNEL_DATA, 0x4ea00000 | PTE_KERNEL_DATA,
    0x4ec00000 | PTE_KERNEL_DATA, 0x4ee00000 | PTE_KERNEL_DATA,
    0x4f000000 | PTE_KERNEL_DATA, 0x4f200000 | PTE_KERNEL_DATA,
    0x4f400000 | PTE_KERNEL_DATA, 0x4f600000 | PTE_KERNEL_DATA,
    0x4f800000 | PTE_KERNEL_DATA, 0x4fa00000 | PTE_KERNEL_DATA,
    0x4fc00000 | PTE_KERNEL_DATA, 0x4fe00000 | PTE_KERNEL_DATA,
    0x50000000 | PTE_KERNEL_DATA, 0x50200000 | PTE_KERNEL_DATA,
    0x50400000 | PTE_KERNEL_DATA, 0x50600000 | PTE_KERNEL_DATA,
    0x50800000 | PTE_KERNEL_DATA, 0x50a00000 | PTE_KERNEL_DATA,
    0x50c00000 | PTE_KERNEL_DATA, 0x50e00000 | PTE_KERNEL_DATA,
    0x51000000 | PTE_KERNEL_DATA, 0x51200000 | PTE_KERNEL_DATA,
    0x51400000 | PTE_KERNEL_DATA, 0x51600000 | PTE_KERNEL_DATA,
    0x51800000 | PTE_KERNEL_DATA, 0x51a00000 | PTE_KERNEL_DATA,
    0x51c00000 | PTE_KERNEL_DATA, 0x51e00000 | PTE_KERNEL_DATA,
    0x52000000 | PTE_KERNEL_DATA, 0x52200000 | PTE_KERNEL_DATA,
    0x52400000 | PTE_KERNEL_DATA, 0x52600000 | PTE_KERNEL_DATA,
    0x52800000 | PTE_KERNEL_DATA, 0x52a00000 | PTE_KERNEL_DATA,
    0x52c00000 | PTE_KERNEL_DATA, 0x52e00000 | PTE_KERNEL_DATA,
    0x53000000 | PTE_KERNEL_DATA, 0x53200000 | PTE_KERNEL_DATA,
    0x53400000 | PTE_KERNEL_DATA, 0x53600000 | PTE_KERNEL_DATA,
    0x53800000 | PTE_KERNEL_DATA, 0x53a00000 | PTE_KERNEL_DATA,
    0x53c00000 | PTE_KERNEL_DATA, 0x53e00000 | PTE_KERNEL_DATA,
    0x54000000 | PTE_KERNEL_DATA, 0x54200000 | PTE_KERNEL_DATA,
    0x54400000 | PTE_KERNEL_DATA, 0x54600000 | PTE_KERNEL_DATA,
    0x54800000 | PTE_KERNEL_DATA, 0x54a00000 | PTE_KERNEL_DATA,
    0x54c00000 | PTE_KERNEL_DATA, 0x54e00000 | PTE_KERNEL_DATA,
    0x55000000 | PTE_KERNEL_DATA, 0x55200000 | PTE_KERNEL_DATA,
    0x55400000 | PTE_KERNEL_DATA, 0x55600000 | PTE_KERNEL_DATA,
    0x55800000 | PTE_KERNEL_DATA, 0x55a00000 | PTE_KERNEL_DATA,
    0x55c00000 | PTE_KERNEL_DATA, 0x55e00000 | PTE_KERNEL_DATA,
    0x56000000 | PTE_KERNEL_DATA, 0x56200000 | PTE_KERNEL_DATA,
    0x56400000 | PTE_KERNEL_DATA, 0x56600000 | PTE_KERNEL_DATA,
    0x56800000 | PTE_KERNEL_DATA, 0x56a00000 | PTE_KERNEL_DATA,
    0x56c00000 | PTE_KERNEL_DATA, 0x56e00000 | PTE_KERNEL_DATA,
    0x57000000 | PTE_KERNEL_DATA, 0x57200000 | PTE_KERNEL_DATA,
    0x57400000 | PTE_KERNEL_DATA, 0x57600000 | PTE_KERNEL_DATA,
    0x57800000 | PTE_KERNEL_DATA, 0x57a00000 | PTE_KERNEL_DATA,
    0x57c00000 | PTE_KERNEL_DATA, 0x57e00000 | PTE_KERNEL_DATA,
    0x58000000 | PTE_KERNEL_DATA, 0x58200000 | PTE_KERNEL_DATA,
    0x58400000 | PTE_KERNEL_DATA, 0x58600000 | PTE_KERNEL_DATA,
    0x58800000 | PTE_KERNEL_DATA, 0x58a00000 | PTE_KERNEL_DATA,
    0x58c00000 | PTE_KERNEL_DATA, 0x58e00000 | PTE_KERNEL_DATA,
    0x59000000 | PTE_KERNEL_DATA, 0x59200000 | PTE_KERNEL_DATA,
    0x59400000 | PTE_KERNEL_DATA, 0x59600000 | PTE_KERNEL_DATA,
    0x59800000 | PTE_KERNEL_DATA, 0x59a00000 | PTE_KERNEL_DATA,
    0x59c00000 | PTE_KERNEL_DATA, 0x59e00000 | PTE_KERNEL_DATA,
    0x5a000000 | PTE_KERNEL_DATA, 0x5a200000 | PTE_KERNEL_DATA,
    0x5a400000 | PTE_KERNEL_DATA, 0x5a600000 | PTE_KERNEL_DATA,
    0x5a800000 | PTE_KERNEL_DATA, 0x5aa00000 | PTE_KERNEL_DATA,
    0x5ac00000 | PTE_KERNEL_DATA, 0x5ae00000 | PTE_KERNEL_DATA,
    0x5b000000 | PTE_KERNEL_DATA, 0x5b200000 | PTE_KERNEL_DATA,
    0x5b400000 | PTE_KERNEL_DATA, 0x5b600000 | PTE_KERNEL_DATA,
    0x5b800000 | PTE_KERNEL_DATA, 0x5ba00000 | PTE_KERNEL_DATA,
    0x5bc00000 | PTE_KERNEL_DATA, 0x5be00000 | PTE_KERNEL_DATA,
    0x5c000000 | PTE_KERNEL_DATA, 0x5c200000 | PTE_KERNEL_DATA,
    0x5c400000 | PTE_KERNEL_DATA, 0x5c600000 | PTE_KERNEL_DATA,
    0x5c800000 | PTE_KERNEL_DATA, 0x5ca00000 | PTE_KERNEL_DATA,
    0x5cc00000 | PTE_KERNEL_DATA, 0x5ce00000 | PTE_KERNEL_DATA,
    0x5d000000 | PTE_KERNEL_DATA, 0x5d200000 | PTE_KERNEL_DATA,
    0x5d400000 | PTE_KERNEL_DATA, 0x5d600000 | PTE_KERNEL_DATA,
    0x5d800000 | PTE_KERNEL_DATA, 0x5da00000 | PTE_KERNEL_DATA,
    0x5dc00000 | PTE_KERNEL_DATA, 0x5de00000 | PTE_KERNEL_DATA,
    0x5e000000 | PTE_KERNEL_DATA, 0x5e200000 | PTE_KERNEL_DATA,
    0x5e400000 | PTE_KERNEL_DATA, 0x5e600000 | PTE_KERNEL_DATA,
    0x5e800000 | PTE_KERNEL_DATA, 0x5ea00000 | PTE_KERNEL_DATA,
    0x5ec00000 | PTE_KERNEL_DATA, 0x5ee00000 | PTE_KERNEL_DATA,
    0x5f000000 | PTE_KERNEL_DATA, 0x5f200000 | PTE_KERNEL_DATA,
    0x5f400000 | PTE_KERNEL_DATA, 0x5f600000 | PTE_KERNEL_DATA,
    0x5f800000 | PTE_KERNEL_DATA, 0x5fa00000 | PTE_KERNEL_DATA,
    0x5fc00000 | PTE_KERNEL_DATA, 0x5fe00000 | PTE_KERNEL_DATA,
    0x60000000 | PTE_KERNEL_DATA, 0x60200000 | PTE_KERNEL_DATA,
    0x60400000 | PTE_KERNEL_DATA, 0x60600000 | PTE_KERNEL_DATA,
    0x60800000 | PTE_KERNEL_DATA, 0x60a00000 | PTE_KERNEL_DATA,
    0x60c00000 | PTE_KERNEL_DATA, 0x60e00000 | PTE_KERNEL_DATA,
    0x61000000 | PTE_KERNEL_DATA, 0x61200000 | PTE_KERNEL_DATA,
    0x61400000 | PTE_KERNEL_DATA, 0x61600000 | PTE_KERNEL_DATA,
    0x61800000 | PTE_KERNEL_DATA, 0x61a00000 | PTE_KERNEL_DATA,
    0x61c00000 | PTE_KERNEL_DATA, 0x61e00000 | PTE_KERNEL_DATA,
    0x62000000 | PTE_KERNEL_DATA, 0x62200000 | PTE_KERNEL_DATA,
    0x62400000 | PTE_KERNEL_DATA, 0x62600000 | PTE_KERNEL_DATA,
    0x62800000 | PTE_KERNEL_DATA, 0x62a00000 | PTE_KERNEL_DATA,
    0x62c00000 | PTE_KERNEL_DATA, 0x62e00000 | PTE_KERNEL_DATA,
    0x63000000 | PTE_KERNEL_DATA, 0x63200000 | PTE_KERNEL_DATA,
    0x63400000 | PTE_KERNEL_DATA, 0x63600000 | PTE_KERNEL_DATA,
    0x63800000 | PTE_KERNEL_DATA, 0x63a00000 | PTE_KERNEL_DATA,
    0x63c00000 | PTE_KERNEL_DATA, 0x63e00000 | PTE_KERNEL_DATA,
    0x64000000 | PTE_KERNEL_DATA, 0x64200000 | PTE_KERNEL_DATA,
    0x64400000 | PTE_KERNEL_DATA, 0x64600000 | PTE_KERNEL_DATA,
    0x64800000 | PTE_KERNEL_DATA, 0x64a00000 | PTE_KERNEL_DATA,
    0x64c00000 | PTE_KERNEL_DATA, 0x64e00000 | PTE_KERNEL_DATA,
    0x65000000 | PTE_KERNEL_DATA, 0x65200000 | PTE_KERNEL_DATA,
    0x65400000 | PTE_KERNEL_DATA, 0x65600000 | PTE_KERNEL_DATA,
    0x65800000 | PTE_KERNEL_DATA, 0x65a00000 | PTE_KERNEL_DATA,
    0x65c00000 | PTE_KERNEL_DATA, 0x65e00000 | PTE_KERNEL_DATA,
    0x66000000 | PTE_KERNEL_DATA, 0x66200000 | PTE_KERNEL_DATA,
    0x66400000 | PTE_KERNEL_DATA, 0x66600000 | PTE_KERNEL_DATA,
    0x66800000 | PTE_KERNEL_DATA, 0x66a00000 | PTE_KERNEL_DATA,
    0x66c00000 | PTE_KERNEL_DATA, 0x66e00000 | PTE_KERNEL_DATA,
    0x67000000 | PTE_KERNEL_DATA, 0x67200000 | PTE_KERNEL_DATA,
    0x67400000 | PTE_KERNEL_DATA, 0x67600000 | PTE_KERNEL_DATA,
    0x67800000 | PTE_KERNEL_DATA, 0x67a00000 | PTE_KERNEL_DATA,
    0x67c00000 | PTE_KERNEL_DATA, 0x67e00000 | PTE_KERNEL_DATA,
    0x68000000 | PTE_KERNEL_DATA, 0x68200000 | PTE_KERNEL_DATA,
    0x68400000 | PTE_KERNEL_DATA, 0x68600000 | PTE_KERNEL_DATA,
    0x68800000 | PTE_KERNEL_DATA, 0x68a00000 | PTE_KERNEL_DATA,
    0x68c00000 | PTE_KERNEL_DATA, 0x68e00000 | PTE_KERNEL_DATA,
    0x69000000 | PTE_KERNEL_DATA, 0x69200000 | PTE_KERNEL_DATA,
    0x69400000 | PTE_KERNEL_DATA, 0x69600000 | PTE_KERNEL_DATA,
    0x69800000 | PTE_KERNEL_DATA, 0x69a00000 | PTE_KERNEL_DATA,
    0x69c00000 | PTE_KERNEL_DATA, 0x69e00000 | PTE_KERNEL_DATA,
    0x6a000000 | PTE_KERNEL_DATA, 0x6a200000 | PTE_KERNEL_DATA,
    0x6a400000 | PTE_KERNEL_DATA, 0x6a600000 | PTE_KERNEL_DATA,
    0x6a800000 | PTE_KERNEL_DATA, 0x6aa00000 | PTE_KERNEL_DATA,
    0x6ac00000 | PTE_KERNEL_DATA, 0x6ae00000 | PTE_KERNEL_DATA,
    0x6b000000 | PTE_KERNEL_DATA, 0x6b200000 | PTE_KERNEL_DATA,
    0x6b400000 | PTE_KERNEL_DATA, 0x6b600000 | PTE_KERNEL_DATA,
    0x6b800000 | PTE_KERNEL_DATA, 0x6ba00000 | PTE_KERNEL_DATA,
    0x6bc00000 | PTE_KERNEL_DATA, 0x6be00000 | PTE_KERNEL_DATA,
    0x6c000000 | PTE_KERNEL_DATA, 0x6c200000 | PTE_KERNEL_DATA,
    0x6c400000 | PTE_KERNEL_DATA, 0x6c600000 | PTE_KERNEL_DATA,
    0x6c800000 | PTE_KERNEL_DATA, 0x6ca00000 | PTE_KERNEL_DATA,
    0x6cc00000 | PTE_KERNEL_DATA, 0x6ce00000 | PTE_KERNEL_DATA,
    0x6d000000 | PTE_KERNEL_DATA, 0x6d200000 | PTE_KERNEL_DATA,
    0x6d400000 | PTE_KERNEL_DATA, 0x6d600000 | PTE_KERNEL_DATA,
    0x6d800000 | PTE_KERNEL_DATA, 0x6da00000 | PTE_KERNEL_DATA,
    0x6dc00000 | PTE_KERNEL_DATA, 0x6de00000 | PTE_KERNEL_DATA,
    0x6e000000 | PTE_KERNEL_DATA, 0x6e200000 | PTE_KERNEL_DATA,
    0x6e400000 | PTE_KERNEL_DATA, 0x6e600000 | PTE_KERNEL_DATA,
    0x6e800000 | PTE_KERNEL_DATA, 0x6ea00000 | PTE_KERNEL_DATA,
    0x6ec00000 | PTE_KERNEL_DATA, 0x6ee00000 | PTE_KERNEL_DATA,
    0x6f000000 | PTE_KERNEL_DATA, 0x6f200000 | PTE_KERNEL_DATA,
    0x6f400000 | PTE_KERNEL_DATA, 0x6f600000 | PTE_KERNEL_DATA,
    0x6f800000 | PTE_KERNEL_DATA, 0x6fa00000 | PTE_KERNEL_DATA,
    0x6fc00000 | PTE_KERNEL_DATA, 0x6fe00000 | PTE_KERNEL_DATA,
    0x70000000 | PTE_KERNEL_DATA, 0x70200000 | PTE_KERNEL_DATA,
    0x70400000 | PTE_KERNEL_DATA, 0x70600000 | PTE_KERNEL_DATA,
    0x70800000 | PTE_KERNEL_DATA, 0x70a00000 | PTE_KERNEL_DATA,
    0x70c00000 | PTE_KERNEL_DATA, 0x70e00000 | PTE_KERNEL_DATA,
    0x71000000 | PTE_KERNEL_DATA, 0x71200000 | PTE_KERNEL_DATA,
    0x71400000 | PTE_KERNEL_DATA, 0x71600000 | PTE_KERNEL_DATA,
    0x71800000 | PTE_KERNEL_DATA, 0x71a00000 | PTE_KERNEL_DATA,
    0x71c00000 | PTE_KERNEL_DATA, 0x71e00000 | PTE_KERNEL_DATA,
    0x72000000 | PTE_KERNEL_DATA, 0x72200000 | PTE_KERNEL_DATA,
    0x72400000 | PTE_KERNEL_DATA, 0x72600000 | PTE_KERNEL_DATA,
    0x72800000 | PTE_KERNEL_DATA, 0x72a00000 | PTE_KERNEL_DATA,
    0x72c00000 | PTE_KERNEL_DATA, 0x72e00000 | PTE_KERNEL_DATA,
    0x73000000 | PTE_KERNEL_DATA, 0x73200000 | PTE_KERNEL_DATA,
    0x73400000 | PTE_KERNEL_DATA, 0x73600000 | PTE_KERNEL_DATA,
    0x73800000 | PTE_KERNEL_DATA, 0x73a00000 | PTE_KERNEL_DATA,
    0x73c00000 | PTE_KERNEL_DATA, 0x73e00000 | PTE_KERNEL_DATA,
    0x74000000 | PTE_KERNEL_DATA, 0x74200000 | PTE_KERNEL_DATA,
    0x74400000 | PTE_KERNEL_DATA, 0x74600000 | PTE_KERNEL_DATA,
    0x74800000 | PTE_KERNEL_DATA, 0x74a00000 | PTE_KERNEL_DATA,
    0x74c00000 | PTE_KERNEL_DATA, 0x74e00000 | PTE_KERNEL_DATA,
    0x75000000 | PTE_KERNEL_DATA, 0x75200000 | PTE_KERNEL_DATA,
    0x75400000 | PTE_KERNEL_DATA, 0x75600000 | PTE_KERNEL_DATA,
    0x75800000 | PTE_KERNEL_DATA, 0x75a00000 | PTE_KERNEL_DATA,
    0x75c00000 | PTE_KERNEL_DATA, 0x75e00000 | PTE_KERNEL_DATA,
    0x76000000 | PTE_KERNEL_DATA, 0x76200000 | PTE_KERNEL_DATA,
    0x76400000 | PTE_KERNEL_DATA, 0x76600000 | PTE_KERNEL_DATA,
    0x76800000 | PTE_KERNEL_DATA, 0x76a00000 | PTE_KERNEL_DATA,
    0x76c00000 | PTE_KERNEL_DATA, 0x76e00000 | PTE_KERNEL_DATA,
    0x77000000 | PTE_KERNEL_DATA, 0x77200000 | PTE_KERNEL_DATA,
    0x77400000 | PTE_KERNEL_DATA, 0x77600000 | PTE_KERNEL_DATA,
    0x77800000 | PTE_KERNEL_DATA, 0x77a00000 | PTE_KERNEL_DATA,
    0x77c00000 | PTE_KERNEL_DATA, 0x77e00000 | PTE_KERNEL_DATA,
    0x78000000 | PTE_KERNEL_DATA, 0x78200000 | PTE_KERNEL_DATA,
    0x78400000 | PTE_KERNEL_DATA, 0x78600000 | PTE_KERNEL_DATA,
    0x78800000 | PTE_KERNEL_DATA, 0x78a00000 | PTE_KERNEL_DATA,
    0x78c00000 | PTE_KERNEL_DATA, 0x78e00000 | PTE_KERNEL_DATA,
    0x79000000 | PTE_KERNEL_DATA, 0x79200000 | PTE_KERNEL_DATA,
    0x79400000 | PTE_KERNEL_DATA, 0x79600000 | PTE_KERNEL_DATA,
    0x79800000 | PTE_KERNEL_DATA, 0x79a00000 | PTE_KERNEL_DATA,
    0x79c00000 | PTE_KERNEL_DATA, 0x79e00000 | PTE_KERNEL_DATA,
    0x7a000000 | PTE_KERNEL_DATA, 0x7a200000 | PTE_KERNEL_DATA,
    0x7a400000 | PTE_KERNEL_DATA, 0x7a600000 | PTE_KERNEL_DATA,
    0x7a800000 | PTE_KERNEL_DATA, 0x7aa00000 | PTE_KERNEL_DATA,
    0x7ac00000 | PTE_KERNEL_DATA, 0x7ae00000 | PTE_KERNEL_DATA,
    0x7b000000 | PTE_KERNEL_DATA, 0x7b200000 | PTE_KERNEL_DATA,
    0x7b400000 | PTE_KERNEL_DATA, 0x7b600000 | PTE_KERNEL_DATA,
    0x7b800000 | PTE_KERNEL_DATA, 0x7ba00000 | PTE_KERNEL_DATA,
    0x7bc00000 | PTE_KERNEL_DATA, 0x7be00000 | PTE_KERNEL_DATA,
    0x7c000000 | PTE_KERNEL_DATA, 0x7c200000 | PTE_KERNEL_DATA,
    0x7c400000 | PTE_KERNEL_DATA, 0x7c600000 | PTE_KERNEL_DATA,
    0x7c800000 | PTE_KERNEL_DATA, 0x7ca00000 | PTE_KERNEL_DATA,
    0x7cc00000 | PTE_KERNEL_DATA, 0x7ce00000 | PTE_KERNEL_DATA,
    0x7d000000 | PTE_KERNEL_DATA, 0x7d200000 | PTE_KERNEL_DATA,
    0x7d400000 | PTE_KERNEL_DATA, 0x7d600000 | PTE_KERNEL_DATA,
    0x7d800000 | PTE_KERNEL_DATA, 0x7da00000 | PTE_KERNEL_DATA,
    0x7dc00000 | PTE_KERNEL_DATA, 0x7de00000 | PTE_KERNEL_DATA,
    0x7e000000 | PTE_KERNEL_DATA, 0x7e200000 | PTE_KERNEL_DATA,
    0x7e400000 | PTE_KERNEL_DATA, 0x7e600000 | PTE_KERNEL_DATA,
    0x7e800000 | PTE_KERNEL_DATA, 0x7ea00000 | PTE_KERNEL_DATA,
    0x7ec00000 | PTE_KERNEL_DATA, 0x7ee00000 | PTE_KERNEL_DATA,
    0x7f000000 | PTE_KERNEL_DATA, 0x7f200000 | PTE_KERNEL_DATA,
    0x7f400000 | PTE_KERNEL_DATA, 0x7f600000 | PTE_KERNEL_DATA,
    0x7f800000 | PTE_KERNEL_DATA, 0x7fa00000 | PTE_KERNEL_DATA,
    0x7fc00000 | PTE_KERNEL_DATA, 0x7fe00000 | PTE_KERNEL_DATA
};

__attribute__((__aligned__(PAGE_SIZE))) PTEntries _kernel_pt_level1 = {
    K2P(_kernel_pt_lv2_dev) + PTE_TABLE,
    K2P(_kernel_pt_lv2_ram) + PTE_TABLE,
};

__attribute__((__aligned__(PAGE_SIZE))) PTEntries kernel_pt_level0 = {
    K2P(_kernel_pt_level1) + PTE_TABLE
};

__attribute__((__aligned__(PAGE_SIZE))) PTEntries invalid_pt = { 0 };