#include <kernel/proc.h>
#include <kernel/mem.h>
#include <kernel/sched.h>
#include <aarch64/mmu.h>
#include <common/list.h>
#include <common/string.h>
#include <kernel/printk.h>
#include <kernel/cpu.h>

Proc root_proc;
PidNode *pid_recycle_list = NULL;  // PID回收链表的头部指针

void kernel_entry();
void proc_entry();

static int pid;
static SpinLock pidlock;
static SpinLock elock;
static SpinLock klock;
#define CR_COUNT 20
static CR cr_resources[CR_COUNT];
static int CR_lock[CR_COUNT];
//回收pid，加入链表
void recycle_pid(int pid) {
    PidNode *node = kalloc(sizeof(PidNode));
    acquire_spinlock(&pidlock);
    node->pid = pid;
    node->next = pid_recycle_list;
    pid_recycle_list = node;
    release_spinlock(&pidlock);
}

//获取pid
int get_pid() {
    if (pid_recycle_list) {
        // 如果回收链表不为空，从链表中取出一个PID
        acquire_spinlock(&pidlock);
        PidNode *node = pid_recycle_list;
        int recycled_pid = node->pid;
        pid_recycle_list = node->next;
        kfree(node);  // 释放链表节点
        release_spinlock(&pidlock);
        return recycled_pid;
    } else {
        // 如果回收链表为空，正常自增PID
        return ++pid;
    }
}

void init_cr_resources() {
    for (int i = 0; i < CR_COUNT; i++) {
        init_sleeplock(&cr_resources[i].lock);  // 初始化自旋锁
        cr_resources[i].priority = 20;       // 设置不同的优先级（例如：1, 2, ..., 20）
        init_list_node(&cr_resources[i].node);   // 初始化链表节点
        cr_resources[i].cid = i;           //  资源的id
    }
}


// init_kproc initializes the kernel process
// NOTE: should call after kinit
void Rinit_kproc(){
    init_spinlock(&pidlock);
    init_spinlock(&elock);
    init_spinlock(&klock);
    init_cr_resources();
    RT_init(&root_proc,1000,40);
    root_proc.parent = &root_proc;
    start_proc(&root_proc, kernel_entry, 123456);
}
void init_kproc()
{
    // TODO:
    // 1. init global resources (e.g. locks, semaphores)
    // 2. init the root_proc (finished)
    // for(int i=0;i<NCPU;i++){
    //     init_spinlock(&plock[i]);
    // }
    init_spinlock(&pidlock);
    init_spinlock(&elock);
    init_spinlock(&klock);
    init_proc(&root_proc);
    root_proc.parent = &root_proc;
    start_proc(&root_proc, kernel_entry, 123456);
}

void RT_init(Proc*p,int period,int exec_time){
    init_proc(p);
    acquire_spinlock(&p->plock);
    p->period = period;
    p->exec_time =exec_time;
    init_sem(&p->task_sleep,0);
    p->current_time =0;
    printk("%d\n",20-period/100);
    p->prim_priority = MAX(20-period/100,0); // 任务原始优先级,RT队列最高优先级为20
    p->true_priority = p->prim_priority; // 任务的真实优先级
    p->execed_time =0;
    p->complete = false;
    release_spinlock(&p->plock);
}

void init_proc(Proc *p)
{
    // TODO:
    // setup the Proc with kstack and pid allocated
    // NOTE: be careful of concurrency
    // 初始化为UNUSED状态
    // ucontext = base - 16 - sizeof(UserContext)
    // kcontext = base - 16 - sizeof(UserContext)-sizeof(KernelContext)
    memset(p,0,sizeof(*p));
    init_spinlock(&p->plock);
    acquire_spinlock(&p->plock);
    p->pid = get_pid();  // 使用get_pid()来获取PID
    init_sem(&p->childexit,0);
    init_pgdir(&p->pgdir);
    init_list_node(&p->children);
    init_list_node(&p->ptnode);
    init_list_node(&p->crnode);
    p->killed = 0;
    p->idle = 0;
    p->wait_time =0;
    p->state = UNUSED;
    p->parent = NULL;
    p->kstack= kalloc_page();
    init_schinfo(&p->schinfo);
    p->kcontext = (KernelContext*)((u64)p->kstack + PAGE_SIZE - 16 - sizeof(KernelContext)-sizeof(UserContext));
    p->ucontext = (UserContext*)  ((u64)p->kstack + PAGE_SIZE - 16 - sizeof(UserContext));
    release_spinlock(&p->plock);
}
Proc *RT_create(int period,int exec_time){
    Proc *p = kalloc(sizeof(Proc));
    RT_init(p,period,exec_time);
    return p;
}
Proc *create_proc()
{
    Proc *p = kalloc(sizeof(Proc));
    init_proc(p);
    return p;
}

void set_parent_to_this(Proc *proc)
{
    // TODO: set the parent of proc to thisproc
    // NOTE: maybe you need to lock the process tree
    // NOTE: it's ensured that the old proc->parent = NULL
    //确保这个函数只使用在未start的进程上
    auto this = thisproc();
    acquire_spinlock(&proc->plock);
    if(proc->parent==NULL){
        proc->parent = thisproc();
        release_spinlock(&proc->plock);
        acquire_spinlock(&this->plock);
        _insert_into_list(&thisproc()->children,&proc->ptnode);
        release_spinlock(&this->plock);
        return;
    }
    PANIC();//只对新创建的进程使用
}

int start_proc(Proc *p, void (*entry)(u64), u64 arg)
{
    // TODO:
    // 1. set the parent to root_proc if NULL
    // 2. setup the kcontext to make the proc start with proc_entry(entry, arg)
    // 3. activate the proc and return its pid
    // NOTE: be careful of concurrency
    if(p->parent ==NULL ){
        
        p->parent = &root_proc;
        acquire_spinlock(&root_proc.plock);
        _insert_into_list(&root_proc.children,&p->ptnode);
        release_spinlock(&root_proc.plock);
        
    }
    p->kcontext->lr = (u64)&proc_entry;
    p->kcontext->x0 = (u64)entry;
    p->kcontext->x1 = (u64) arg;
    int id = p->pid;
    activate_proc(p);

    return id;
}
int start_proc_with_id(Proc *p, void (*entry)(u64), u64 arg,int cid)
{
    // TODO:
    // 1. set the parent to root_proc if NULL
    // 2. setup the kcontext to make the proc start with proc_entry(entry, arg)
    // 3. activate the proc and return its pid
    // NOTE: be careful of concurrency
    if(p->parent ==NULL ){
        
        p->parent = &root_proc;
        acquire_spinlock(&root_proc.plock);
        _insert_into_list(&root_proc.children,&p->ptnode);
        release_spinlock(&root_proc.plock);
        
    }
    p->kcontext->lr = (u64)&proc_entry;
    p->kcontext->x0 = (u64)entry;
    p->kcontext->x1 = (u64) arg;
    int id = p->pid;
    _activate_proc_with_id(p,false,cid);
    return id;
}
int wait(int *exitcode)
{
    auto this = thisproc();

    acquire_spinlock(&this->plock);
    if (_empty_list(&this->children)) {
        release_spinlock(&this->plock);
        return -1;
    }
    release_spinlock(&this->plock);

    bool x = wait_sem(&this->childexit);
    if (x == false) {
        PANIC();
    }

    acquire_spinlock(&this->plock);

    auto node = this->children.next;
    while (node != &this->children) {
        Proc *x = container_of(node, Proc, ptnode);
        node = node->next;
        acquire_spinlock(&x->plock);
        if (x->state == ZOMBIE) {
            recycle_pid(x->pid);
            _detach_from_list(&x->ptnode);
            release_spinlock(&this->plock);


            int id = x->pid;
            *exitcode = x->exitcode;
            release_spinlock(&x->plock);
            kfree_page(x->kstack);
            kfree(x);
            return id;
        }
        release_spinlock(&x->plock);
    }
    release_spinlock(&this->plock);
    ASSERT(1 == 0);
    return -100;
}


void reparent(Proc *this)
{
    if (_empty_list(&this->children))
        return;
    auto node = this->children.next;
    while (node != &this->children) {
        auto x = container_of(node, Proc, ptnode);
        node = node->next;

        acquire_spinlock(&x->plock);
        x->parent = &root_proc;
        release_spinlock(&x->plock);

        _detach_from_list(&x->ptnode);

        acquire_spinlock(&root_proc.plock);
        _insert_into_list(&root_proc.children, &x->ptnode);
        release_spinlock(&root_proc.plock);

        if (is_zombie(x)) {
            post_sem(&root_proc.childexit);
        }
    }
}

NO_RETURN void exit(int code)
{
    acquire_spinlock(&elock);
    auto this = thisproc();

    acquire_spinlock(&this->plock);

    reparent(this);

    
    this->exitcode = code;
    free_pgdir(&this->pgdir);
    if(this->parent !=NULL){
        post_sem(&this->parent->childexit);
    }
    acquire_sched_lock();

    release_spinlock(&this->plock);
    release_spinlock(&elock);

    sched(ZOMBIE);

    PANIC(); // prevent the warning of 'no_return function returns'
}


void dfs_kill(Proc *proc, int pid, bool *killed)
{
    // 遍历当前进程的子进程
    ListNode *node = proc->children.next;
    while (node != &proc->children) {
        Proc *child = container_of(node, Proc, ptnode);
        if(child->pid == pid && is_unused(child))
            return;
        if (child->pid == pid) {
            acquire_spinlock(&child->plock);
            child->killed = 1;  // 找到目标进程，标记为已被杀死
            *killed = true;     // 设置 killed 标志为 true
            release_spinlock(&child->plock);
            alert_proc(child);
            return;
        }
        // 递归处理子进程
        dfs_kill(child, pid, killed);
        node = node->next;  // 移动到下一个子进程
    }
}

int kill(int pid)
{
    // TODO:
    // Set the killed flag of the proc to true and return 0.
    // Return -1 if the pid is invalid (proc not found).
    acquire_spinlock(&klock); // 获取当前 CPU 的自旋锁
    Proc *this_proc = thisproc();       // 获取当前进程

    bool killed = false;  // 初始化 killed 标志为 false
    // 调用 DFS 函数以查找并杀死指定的进程
    dfs_kill(this_proc, pid, &killed);

    release_spinlock(&klock); // 释放自旋锁
    return killed ? 0 : -1;  // 如果 killed 为 true，则返回 0，表示成功；否则返回 -1
}

void init_critical_resource(CR *cr, int priority,int cid) {
    init_sleeplock(&cr->lock);  // 初始化自旋锁
    init_list_node(&cr->node);
    cr->priority = priority;    // 设置优先级
    cr->cid = cid;
}


bool acquire_CR(u64 i) {
    thisproc()->currfuncid=2;
    // 禁用中断，防止在获取临界资源期间发生中断
    //local_irq_save(flags);
    CR_lock[i] = cpuid()+1;
    auto proc = thisproc();
    bool printed = false;
    // 尝试获取自旋锁
    if(!acquire_sleeplock(&cr_resources[i].lock)){
        if(!printed){
            printk("cpu %lld Failed to acquire CR lock. %lld\n",cpuid(),i);
            printed=true;
        }
        return -1;
    }
    // 如果获取成功，更新当前进程的优先级,Ceiling
    if (proc) {
        printk("PID %d,Before priority update: proc->true_priority = %d, cr->priority = %d\n",proc->pid,
            proc->true_priority, cr_resources[i].priority);
        acquire_spinlock(&proc->plock);
        if(cr_resources[i].priority > proc->true_priority)
            proc->true_priority = cr_resources[i].priority;
        else cr_resources[i].priority = proc->true_priority;
        //int current_priority = cr->priority;  // 资源的优先级
        printk("PID %d,After priority update: proc->true_priority = %d, cr->priority = %d\n",proc->pid,
            proc->true_priority, cr_resources[i].priority);
        _insert_into_list(&proc->crnode, &cr_resources[i].node);
        release_spinlock(&proc->plock);
        //set_user_nice(current, current_priority);  // 更新当前进程的优先级
        //spin_unlock(&cr->lock);  // 释放自旋锁
        //local_irq_restore(flags);  // 恢复中断
        return true;  // 获取资源成功
    } else {
        release_sleeplock(&cr_resources[i].lock);  // 释放自旋锁
        //local_irq_restore(flags);  // 恢复中断
        return false;  // 获取资源失败
    }
}

void release_CR(CR *cr) {
    thisproc()->currfuncid=3;
    CR_lock[cr->cid]=0;
    _detach_from_list(&cr->node);
    release_sched_lock();
    printk("CR %d was released by %d at %lld\n",cr->cid,thisproc()->pid,cpuid());
    release_sleeplock(&cr->lock);
    acquire_sched_lock();
}
void trap_return();
Proc* fork() {
    struct Proc* parent_proc = thisproc();
    struct Proc* child_proc = RT_create(parent_proc->period,parent_proc->exec_time);
    if (child_proc == NULL) {
        return NULL;
    }

    struct pgdir* new_pgdir = deep_copy_vm(&parent_proc->pgdir);
    if (new_pgdir == NULL) {
        // Cleanup if page directory copy fails
        kfree(child_proc->kstack);
        acquire_spinlock(&child_proc->plock);
        child_proc->state = UNUSED;
        release_spinlock(&child_proc->plock);
        return NULL;
    }

    child_proc->pgdir = *new_pgdir;
    child_proc->parent = parent_proc;

    memmove(child_proc->ucontext, parent_proc->ucontext, sizeof(*parent_proc->ucontext));

    // The child process should see a return value of 0
    child_proc->ucontext->x[0] = 0;

    acquire_spinlock(&parent_proc->plock);
    _insert_into_list(&parent_proc->children, &child_proc->ptnode);
    release_spinlock(&parent_proc->plock);

    start_proc(child_proc, trap_return, 0);
    return child_proc;
}