#include <kernel/pt.h>
#include <kernel/mem.h>
#include <common/string.h>
#include <aarch64/intrinsic.h>
#include <aarch64/mmu.h>

PTEntriesPtr get_pte(struct pgdir *pgdir, u64 va, bool alloc)
{
    // TODO:
    // Return a pointer to the PTE (Page Table Entry) for virtual address 'va'
    // If the entry not exists (NEEDN'T BE VALID), allocate it if alloc=true, or return NULL if false.
    // THIS ROUTINUE GETS THE PTE, NOT THE PAGE DESCRIBED BY PTE.
    PTEntriesPtr pgtbl = pgdir->pt;

    // 如果一级页表不存在且允许分配，分配一级页表
    if (!pgtbl && alloc) {
        pgtbl = (PTEntriesPtr)kalloc_page();
        memset(pgtbl, 0, PAGE_SIZE);
        pgdir->pt = pgtbl;
    }

    if (!pgtbl) return NULL;

    // 获取虚拟地址的各级索引
    int idx[] = {VA_PART0(va), VA_PART1(va), VA_PART2(va), VA_PART3(va)};


    // 遍历第1到第3级页表
    for (int level = 0; level < 3; level++) {
        PTEntry *pte = &pgtbl[idx[level]];

        // 如果当前页表项无效，并且允许分配新页表
        if (!(*pte & PTE_VALID)) {
            if (!alloc) return NULL;

            // 分配下一级页表
            PTEntriesPtr new_table = (PTEntriesPtr)kalloc_page();
            memset(new_table, 0, PAGE_SIZE);

            // 将新页表的物理地址存入页表项 (需要将虚拟地址转换为物理地址)
            *pte = (u64)K2P(new_table) | PTE_TABLE | PTE_VALID;
        }

        // 获取下一级页表的内核虚拟地址 (从物理地址转换为虚拟地址)
        pgtbl = (PTEntriesPtr)P2K((PTE_ADDRESS(*pte)));
    }

    // 返回最终第四级页表项的地址
    return &pgtbl[idx[3]];
}

void init_pgdir(struct pgdir *pgdir)
{
    pgdir->pt = NULL;
}

    // 定义一个递归函数来释放页表
void free_pagetable(PTEntriesPtr pt, int level) {
    // level最大为3，超过这个层次的页表项是指向实际的物理页，不再递归
    if (level == 4) return;

    for (int i = 0; i < N_PTE_PER_TABLE; i++) {
        PTEntry pte = pt[i];

        if (pte & PTE_VALID) {
            // 如果是页表项（指向下一级页表）
            if ((pte & PTE_TABLE) == PTE_TABLE) {
                // 获取下一级页表的地址
                PTEntriesPtr next_pt = (PTEntriesPtr)P2K(PTE_ADDRESS(pte));
                free_pagetable(next_pt, level + 1);  // 递归释放下一级页表

                // 释放当前级别页表项所指向的下一级页表
                if(level<=2)
                    kfree_page((void *)next_pt);
            }
            // 如果是叶子节点（即指向实际物理页），不进行释放
        }
    }
}


void free_pgdir(struct pgdir *pgdir)
{
    if (!pgdir->pt) return;
    free_pagetable(pgdir->pt, 0);
    // 释放顶级页表
    kfree_page((void *)pgdir->pt);
    pgdir->pt = NULL;
}

void attach_pgdir(struct pgdir *pgdir)
{
    extern PTEntries invalid_pt;
    if (pgdir->pt)
        arch_set_ttbr0(K2P(pgdir->pt));
    else
        arch_set_ttbr0(K2P(&invalid_pt));
}


struct pgdir* deep_copy_vm(struct pgdir* src_pgdir) {
    // Allocate memory for the new page directory and initialize it
    struct pgdir* dest_pgdir = kalloc(sizeof(struct pgdir));
    if (!dest_pgdir) { // Check if allocation was successful
        return NULL;   // Return NULL if allocation failed
    }
    init_pgdir(dest_pgdir); // Initialize the new page directory

    // Traverse the source page directory (Level 0 table)
    for (int l0_idx = 0; l0_idx < N_PTE_PER_TABLE; l0_idx++) {
        if (!(src_pgdir->pt[l0_idx] & PTE_VALID)) { 
            continue; // Skip invalid entries
        }

        ASSERT(src_pgdir->pt[l0_idx] & PTE_TABLE); // Ensure it's a table entry
        PTEntriesPtr l1_table = (PTEntriesPtr)P2K(PTE_ADDRESS(src_pgdir->pt[l0_idx]));

        // Traverse the Level 1 table
        for (int l1_idx = 0; l1_idx < N_PTE_PER_TABLE; l1_idx++) {
            if (!(l1_table[l1_idx] & PTE_VALID)) {
                continue; // Skip invalid entries
            }

            ASSERT(l1_table[l1_idx] & PTE_TABLE); // Ensure it's a table entry
            PTEntriesPtr l2_table = (PTEntriesPtr)P2K(PTE_ADDRESS(l1_table[l1_idx]));

            // Traverse the Level 2 table
            for (int l2_idx = 0; l2_idx < N_PTE_PER_TABLE; l2_idx++) {
                if (!(l2_table[l2_idx] & PTE_VALID)) {
                    continue; // Skip invalid entries
                }

                ASSERT(l2_table[l2_idx] & PTE_TABLE); // Ensure it's a table entry
                PTEntriesPtr l3_table = (PTEntriesPtr)P2K(PTE_ADDRESS(l2_table[l2_idx]));

                // Traverse the Level 3 table (points to actual pages)
                for (int l3_idx = 0; l3_idx < N_PTE_PER_TABLE; l3_idx++) {
                    if (!(l3_table[l3_idx] & PTE_VALID)) {
                        continue; // Skip invalid entries
                    }

                    // Validate that this is a valid page and not a table entry
                    ASSERT(l3_table[l3_idx] & PTE_PAGE);
                    ASSERT(l3_table[l3_idx] & PTE_USER); // Ensure it's a user page
                    ASSERT(l3_table[l3_idx] & PTE_NORMAL);

                    // Calculate the virtual address based on table indices
                    u64 virtual_address = ((u64)l0_idx << (12 + 9 * 3)) |
                                          ((u64)l1_idx << (12 + 9 * 2)) |
                                          ((u64)l2_idx << (12 + 9)) |
                                          ((u64)l3_idx << 12);

                    // Extract the physical address of the page
                    u64 physical_address = PTE_ADDRESS(l3_table[l3_idx]);

                    // Allocate a new physical page for the destination
                    void* new_page = kalloc_page();
                    ASSERT(new_page); // Ensure allocation was successful

                    // Copy the contents of the source page to the new page
                    memmove(new_page, (void*)P2K(physical_address), PAGE_SIZE);

                    // Map the new page in the destination page directory
                    PTEntriesPtr dest_pte = get_pte(dest_pgdir, virtual_address, true);
                    *dest_pte = K2P(new_page) | PTE_USER_DATA; // Set flags for user access
                }
            }
        }
    }

    // Return the new page directory after successful copy
    return dest_pgdir;
}